@@ -20,6 +20,7 @@
 
 #ifndef __engine_h__
 #define __engine_h__
+#include "config.h"
 
 typedef void *handle_t;
 
@@ -29,14 +30,25 @@ typedef enum
   GESTURE_EVENT
 } event_type_t;
 
-typedef void (*recognize_callback_t) (ClutterActor *target_actor,
-                event_type_t event_type, GSList *event_list,
+#define IN_MT_GESTURE (1<<0)
+
+#define DEFAULT_TOUCH_HOLD_INTERVAL 800
+#define DEFAULT_TOUCH_HOLD_RADIUS   20
+
+typedef gboolean (*recognize_callback_t) (ClutterActor *target_actor,
+                event_type_t event_type, guint flag, GSList *event_list,
                 gint number, void *data);
 
 handle_t engine_init(recognize_callback_t cb, void *data);
 void engine_shutdown(handle_t handle);
 
 gint set_gesture_mask(handle_t handle, ClutterActor *target_actor, guint mask);
+gint get_gesture_mask(handle_t handle, ClutterActor *target_actor, guint *mask);
+gboolean set_hold_timeout(handle_t handle, ClutterActor *target_actor, guint interval);
+gboolean get_hold_timeout(handle_t handle, ClutterActor *target_actor, guint *interval);
+gboolean set_hold_radius(handle_t handle, ClutterActor *target_actor, guint radius);
+gboolean get_hold_radius(handle_t handle, ClutterActor *target_actor, guint *radius);
 gboolean fill_event(handle_t handle, ClutterActor *target_actor,
                         ClutterEvent *event);
+
 #endif
